/* $Id: wtb3.h 7220 2024-01-09 20:15:55Z aron $ */

#ifndef WELLTYPE_WTB3_H
#define WELLTYPE_WTB3_H

#include "config.h"
#include <stddef.h>

/*----------------------------------------------------------------*/
/* types IDs, type masks, and type flags */

#define WTB3TF_REF              0x80000000ul
#define WTB3TF_MUTABLE          0x40000000ul /* function argument type modifier: able to modify object content */
#define WTB3TF_SEQ              0x20000000ul
#define WTB3TF_LOCAL            0x10000000ul /* type from the local type vector (not resolved type) */
#define WTB3TF_USERTYPE         0x00800000ul
#define WTB3TM_USERTYPE         0x00fffffful
#define WTB3TM_USERTYPEID       0x007ffffful
#define WTB3TM_BASICTYPE        0x0000001ful
#define WTB3TM_USEDBITS         0xf0fffffful
#define WTB3TM_UNUSEDBITS       (~WTB3TM_USEDBITS)
#define WTB3T_VOID              0
#define WTB3T_BOOL              1
#define WTB3T_I32               2
#define WTB3T_U32               3
#define WTB3T_I64               4
#define WTB3T_U64               5
#define WTB3T_F32               6
#define WTB3T_F64               7
#define WTB3T_CHAR              8
#define WTB3T_UNUSED9           9
#define WTB3T_STRING            10
#define WTB3T_UNUSED11          11
#define WTB3T_HANDLE            12
#define WTB3T_PROGRAM           13
#define WTB3T_UNUSED14          14
#define WTB3T_UNUSED15          15
#define WTB3T_FILE              16
#define WTB3T_DIR               17
#define WTB3T_UNUSED18          18
#define WTB3T_I8                19
#define WTB3T_U8                20
#define WTB3T_I16               21
#define WTB3T_U16               22
#define WTB3T_UNUSED23          23
#define WTB3T_UNUSED24          24
#define WTB3T_UNUSED25          25
#define WTB3T_UNUSED26          26
#define WTB3T_UNUSED27          27
#define WTB3T_UNUSED28          28
#define WTB3T_UNUSED29          29
#define WTB3T_UNUSED30          30
#define WTB3T_UNUSED31          31

/*----------------------------------------------------------------*/
/* binary types */

#define WTB3MAG0                'W'
#define WTB3MAG1                'T'
#define WTB3MAG2                'B'
#define WTB3MAG3                '\x03'

#define WTB3V_CURRENT           1

#define WTB3_NIDENT             16 /* size of 'b_ident' */

#define WTB3I_MAGIC0            0  /* value is WTB3MAG0 */
#define WTB3I_MAGIC1            1  /* value is WTB3MAG1 */
#define WTB3I_MAGIC2            2  /* value is WTB3MAG2 */
#define WTB3I_MAGIC3            3  /* value is WTB3MAG3 */
#define WTB3I_VERSION           4  /* WTB3V_CURRENT */
#define WTB3I_BYTEORDER         5  /* byte order (WTBO_*) */
/*#define WTB3I_                6  */ /* MBZ */
/*#define WTB3I_                7  */ /* MBZ */
#define WTB3I_COMPVER0          8  /* compiler version (major) */
#define WTB3I_COMPVER1          9  /* compiler version (minor) */
#define WTB3I_COMPVER2          10 /* compiler version (patch) */
#define WTB3I_COMPVER3          11 /* MBZ since v0.9.5; used to be 4th component of the compiler version */
/*#define WTB3I_                12 */ /* MBZ */
/*#define WTB3I_                13 */ /* MBZ */
/*#define WTB3I_                14 */ /* MBZ */
/*#define WTB3I_                15 */ /* MBZ */

struct wtb3_hdr
{
	wt_u8  b_ident[WTB3_NIDENT]; /* identification (item indices are WTB3I_*) */
	wt_u32 b_size;               /* number of octets to allocate */
	wt_u16 b_shnum;              /* number of sections */
	wt_u16 b_thnum;              /* number of threads */
	wt_u32 b_shoff;              /* absolute offset to section header */
	wt_u32 b_thoff;              /* absolute offset to thread header */
	wt_u32 b_ctor;               /* PC of constructor code */
	wt_u32 b_dtor;               /* PC of destructor code */
};

#define WTB3ST_UNUSED           0x00 /* skip this entry */
#define WTB3ST_CODE             0x01 /* executable program code (at least one; no BOS) */
#define WTB3ST_DATA             0x02 /* data (if BOS exists, link to WTB3ST_BOS) */
#define WTB3ST_BOS              0x03 /* byte order swap (linked to WTB3ST_DATA; 'sh_addr' is the same) */
#define WTB3ST_IMPORT           0x05 /* import symbol table */
#define WTB3ST_EXPORT           0x06 /* export symbol table */
#define WTB3ST_METADATA         0x07 /* metadata table */
#define WTB3ST_INTEGRITY        0x08 /* integrity (hash algorithm and digest) */
#define WTB3ST_DEBUGLABEL       0xB0 /* map addresses to labels */
#define WTB3ST_CUSTOMLO         0xC0 /* -- */
#define WTB3ST_CUSTOMHI         0xFF /* -- */

#define WTB3SF_CLEAR            0x01 /* if set, clear data on 'sh_addr' */
#define WTB3SF_LOAD             0x02 /* if set, load to 'sh_addr' */
#define WTB3SF_TYPEVEC          0x04 /* is set, import type vector starts on 'sh_addr' (at most one) */

struct wtb3_shdr
{
	wt_u8  sh_type; /* section type (WTB3ST_*) */
	wt_u8  sh_flag; /* section flags (WTB3SF_*) */
	wt_u16 sh_link; /* link to another section */
	wt_u32 sh_offs; /* absolute offset to data */
	wt_u32 sh_size; /* data size in octets */
	wt_u32 sh_addr; /* relative to program base address */
};

#define WTB3_2CC(_hi_,_lo_)     (((_hi_)<<8) | (_lo_))
#define WTB3_4CC(hh,hl,lh,ll)   (((hh)<<24) | ((hl)<<16) | ((lh)<<8) | (ll))

#define WTB3TT_USER             WTB3_2CC('S', 'R') /* user thread (exactly once, no WTB3TF_EXECUTE) */
#define WTB3TT_TASK             WTB3_2CC('T', 'K') /* user defined task (with WTB3TF_EXECUTE) */

#define WTB3TF_EXECUTE          0x0001 /* start thread automatically */

struct wtb3_thdr
{
	wt_u16 th_type;   /* thread type (WTB3TT_*) */
	wt_u16 th_flag;   /* thread flags (WTB3TF_*) */
	wt_u32 th_cssize; /* call stack size */
	wt_u32 th_dssize; /* data stack size */
	wt_u32 th_pcbase; /* start PC */
};

#define WTB3BOS_NOOP            0x00 /* no operation */
#define WTB3BOS_NOSWAP_8        0x01 /* no swap, octet count in 8-bit constant */
#define WTB3BOS_NOSWAP_16       0x02 /* no swap, octet count in 16-bit constant */
#define WTB3BOS_NOSWAP_32       0x03 /* no swap, octet count in 32-bit constant */
#define WTB3BOS_NOSWAP_64       0x04 /* no swap, octet count in 64-bit constant */
#define WTB3BOS_SWAP16_8        0x11 /* swap 16-bit fields, field count in 8-bit constant */
#define WTB3BOS_SWAP16_16       0x12 /* swap 16-bit fields, field count in 16-bit constant */
#define WTB3BOS_SWAP16_32       0x13 /* swap 16-bit fields, field count in 32-bit constant */
#define WTB3BOS_SWAP16_64       0x14 /* swap 16-bit fields, field count in 64-bit constant */
#define WTB3BOS_SWAP32_8        0x21 /* swap 32-bit fields, field count in 8-bit constant */
#define WTB3BOS_SWAP32_16       0x22 /* swap 32-bit fields, field count in 16-bit constant */
#define WTB3BOS_SWAP32_32       0x23 /* swap 32-bit fields, field count in 32-bit constant */
#define WTB3BOS_SWAP32_64       0x24 /* swap 32-bit fields, field count in 64-bit constant */
#define WTB3BOS_SWAP64_8        0x31 /* swap 64-bit fields, field count in 8-bit constant */
#define WTB3BOS_SWAP64_16       0x32 /* swap 64-bit fields, field count in 16-bit constant */
#define WTB3BOS_SWAP64_32       0x33 /* swap 64-bit fields, field count in 32-bit constant */
#define WTB3BOS_SWAP64_64       0x34 /* swap 64-bit fields, field count in 64-bit constant */

#define WTB3SYMT_FUNCTION       WTB3_4CC('f', 'u', 'n', 'c')
#define WTB3SYMT_EXCEPTION      WTB3_4CC('e', 'x', 'c', ' ')
#define WTB3SYMT_TYPE           WTB3_4CC('t', 'y', 'p', 'e')
#define WTB3SYMT_DATATYPE       WTB3_4CC('d', 't', 'y', 'p')
#define WTB3SYMT_OPERATOR       WTB3_4CC('o', 'p', ' ', ' ')
#define WTB3SYMT_RECORD         WTB3_4CC('r', 'e', 'c', ' ')
#define WTB3SYMT_PRIVATERECORD  WTB3_4CC('r', 'e', 'c', 'P')
#define WTB3SYMT_LIMITEDRECORD  WTB3_4CC('r', 'e', 'c', 'L')
#define WTB3SYMT_FUNCTIONTYPE   WTB3_4CC('f', 'u', 'n', 'T')
#define WTB3SYMT_ENUM           WTB3_4CC('e', 'n', 'u', 'm')

struct wtb3_symhdr
{
	wt_u32 symcount;  /* number of wtb3_sym */
	wt_u32 reserved;  /* must be zero */
};

#define WTB3SYM_OPASSOC(_inf_)  ((_inf_) & 0xff)
#define WTB3SYM_OPPREC(_inf_)   (((_inf_)>>8) & 0xff)
#define WTB3SYM_CR_OP(_as,_pr)  (((_pr)<<8) | (_as))
#define WTB3SYM_FUNCPC(_inf_)   (_inf_)
#define WTB3SYM_FUNCOFF(_inf_)  (_inf_)
#define WTB3SYM_CR_FUNC(_pc_)   (_pc_)

#define WTB3SYM_OP_MINPREC      3
#define WTB3SYM_OP_MAXPREC      15
#define WTB3SYM_OP_LEFTASSOC    1
#define WTB3SYM_OP_RIGHTASSOC   2

/*
	IMPORTANT!

	In case of WTB3ST_IMPORT and WTB3ST_EXPORT the 'sh_addr' is
	an offset to the type resolution table of the program.
*/

struct wtb3_sym
{
	wt_u32 sym_type;  /* symbol type (WTB3SYMT_*) */
	wt_u32 sym_name;  /* offset to symbol name (relative to wtb3_symhdr) */
	wt_u32 sym_offs;  /* offset to symbol data */
	wt_u32 sym_info;  /* additional argument of the type (WTB3SYM_*) */
};

/*
	Symbol data:

	function:
	functiontype:
		wt_u32 argc;
		wt_u32 argv[];
		wt_u32 retc;
		wt_u32 retv[];
		wt_u8  pure

	record:
		wt_u32 fldc;
		wt_u32 fldv[];
		struct fldnames[];

	datatype:
		wt_u32 consnum;
		wt_u32 defcons;
		{
			string cons_name; -- align offset to 4 after name
			wt_u32 typn;
			wt_u32 typv[];
		} conses [];

	enum:
		wt_u32 enumeratorc;
		string enumeratorv[];

	type:
		wt_u32 type;
*/

/*
	Symbol info:

	function import:
		offset of PFN (relative to program base address)
	function export:
		PC

	operator:
		associativity (WTB3SYM_OPASSOC)
		precedence    (WTB3SYM_OPPREC)

	record:
		type index

	datatype:
		type index

	enum:
		type index

	functiontype:
		type index

	type:
		type index

	exception import:
		storage address of exception global ID
*/

struct wtb3_metahdr
{
	wt_u32 metacount; /* number of wtb3_meta */
	wt_u32 reserved;  /* must be zero */
};

struct wtb3_meta
{
	wt_u32 meta_key_offs; /* offset to key name (zero-terminated string) */
	wt_u32 meta_val_offs; /* offset to value (zero-terminated string) */
};

struct wtb3_dlabelhdr
{
	wt_u32 labelcount; /* number of wtb3_dlabel */
	wt_u32 reserved;   /* must be zero */
};

struct wtb3_dlabel
{
	wt_u32 addr;     /* address in program space */
	wt_u32 lab_offs; /* offset to name (zero-terminated string) */
};

struct wtb3_integhdr
{
	wt_u16 algorithm; /* hash algorithm identifier (WTBHA_*) */
	wt_u16 reserved;  /* must be zero */
};

#endif /* WELLTYPE_WTB3_H */
