#include "config_private.h"
#include "debug.h"

#if WTRUNTIME_DEBUG

#ifdef LOG_OBJECT_ALLOCATIONS

#include <stdlib.h>
#include "klib/kbtree.h"

struct wtr_ptrset* allocated;

#define DEF_SET(name) \
\
struct name##_elem \
{ \
	const name##set_key* key; \
}; \
\
KBTREE_INIT(name, struct name##_elem, name##set_cmp) \
\
struct wtr_##name##set* wtr_##name##set_create(size_t size) \
{ \
	return (struct wtr_##name##set*)kb_init(name, size); \
} \
\
void wtr_##name##set_delete(struct wtr_##name##set* const set) \
{ \
	kb_destroy(name, ((kbtree_t(name)*)set)); \
} \
\
int wtr_##name##set_has(struct wtr_##name##set* const set, const name##set_key* const key) \
{ \
	struct name##_elem elem; \
\
	elem.key = key; \
\
	{ \
		struct name##_elem* const res = kb_getp(name, ((kbtree_t(name)*)set), &elem); \
\
		if(NULL!=res) \
		{ \
			return 1; \
		} \
	} \
\
	return 0; \
} \
\
int wtr_##name##set_add(struct wtr_##name##set* const set, const name##set_key* const key) \
{ \
	struct name##_elem elem; \
\
	elem.key = key; \
\
	{ \
		struct name##_elem* const res = kb_getp(name, ((kbtree_t(name)*)set), &elem); \
\
		if(NULL!=res) \
		{ \
			/* already exists */ \
			return -1; \
		} \
	} \
\
	if(NULL==kb_putp(name, ((kbtree_t(name)*)set), &elem)) \
	{ \
		/* out of memory? */ \
		return -1; \
	} \
\
	return 0; \
} \
\
void wtr_##name##set_remove(struct wtr_##name##set* const set, const name##set_key* const key) \
{ \
	struct name##_elem elem; \
\
	elem.key = (name##set_key*)key; \
\
	kb_delp(name, ((kbtree_t(name)*)set), &elem); \
}

static int ptrcmp(const void* lhs, const void* rhs)
{
	if(((wt_uintptr)lhs)<((wt_uintptr)rhs)) { return -1; }
	if(((wt_uintptr)lhs)>((wt_uintptr)rhs)) { return +1; }
	return 0;
}

typedef void ptrset_key;
#define ptrset_cmp(a, b) (ptrcmp((a).key, (b).key))

DEF_SET(ptr)

#endif /* LOG_OBJECT_ALLOCATIONS */

#ifdef ADD_OBJECT_MAGIC

void wtr_debug_check_all_bytes(void* data, char ch, size_t size, wt_bool clear)
{
	char* chs = (char*)data;

	while(size)
	{
		fflush(stdout);
		assert(ch==*chs);

		if(clear)
		{
			*chs = '\x01';
		}

		++chs;
		--size;
	}
}

#endif /* ADD_OBJECT_MAGIC */

#endif /* WTRUNTIME_DEBUG */
