#include "config_private.h"
#include "loader.h"
#include "loader3.h"
#include "lreader.h"
#include "vm.h"
#include "welltype/wtb3.h"

union wtbx_hdr
{
	struct wtb3_hdr wtb3_hdr;
};

wt_error wtr_loader_sys(struct wtr_vm* const vm,
	struct wtr_program** const program, struct wtr_context* const ctx,
	const void* data, size_t size, wt_uint32_t flags)
{
	struct wtr_lreader lr;

	const union wtbx_hdr* const hdr = (const union wtbx_hdr*)data;

	if(WTB3MAG0==hdr->wtb3_hdr.b_ident[WTB3I_MAGIC0] &&
		WTB3MAG1==hdr->wtb3_hdr.b_ident[WTB3I_MAGIC1] &&
		WTB3MAG2==hdr->wtb3_hdr.b_ident[WTB3I_MAGIC2] &&
		WTB3MAG3==hdr->wtb3_hdr.b_ident[WTB3I_MAGIC3])
	{
		/* appears to be a WTB3 */

		/* use WTB's byteorder here, assuming it is valid -- it will be */
		/* validated later */
		wtr_lreader_init_sys(&lr, vm, data, size,
			hdr->wtb3_hdr.b_ident[WTB3I_BYTEORDER]);

		return wtr_loader3(vm, program, ctx, &lr, flags, &hdr->wtb3_hdr);
	}

	return WTE_BINERR;
}

static wt_error wtr_loader_vm_impl(struct wtr_vm* const vm,
	struct wtr_program** const program, struct wtr_context* const ctx,
	wt_vmptr data, size_t size, wt_uint32_t flags)
{
	struct wtr_lreader lr;

	union wtbx_hdr hdr;

	if(0!=wtr_vma_read(vm, data, &hdr, sizeof(hdr)))
	{
		return WTE_FAULT;
	}

	if(WTB3MAG0==hdr.wtb3_hdr.b_ident[WTB3I_MAGIC0] &&
		WTB3MAG1==hdr.wtb3_hdr.b_ident[WTB3I_MAGIC1] &&
		WTB3MAG2==hdr.wtb3_hdr.b_ident[WTB3I_MAGIC2] &&
		WTB3MAG3==hdr.wtb3_hdr.b_ident[WTB3I_MAGIC3])
	{
		/* appears to be a WTB3 */

		/* use WTB's byteorder here, assuming it is valid -- it will be */
		/* validated later */
		wtr_lreader_init_vm(&lr, vm, data, size,
			hdr.wtb3_hdr.b_ident[WTB3I_BYTEORDER]);

		return wtr_loader3(vm, program, ctx, &lr, flags, &hdr.wtb3_hdr);
	}

	return WTE_BINERR;
}

wt_error wtr_loader_vm(struct wtr_vm* const vm,
	struct wtr_program** const program, struct wtr_context* const ctx,
	wt_vmptr data, size_t size, wt_uint32_t flags)
{
	if(0!=wtr_vma_acquire(vm, data, size))
	{
		return WTE_FAULT;
	}

	const wt_error err = wtr_loader_vm_impl(vm, program, ctx, data, size,
		flags);

	wtr_vma_release(vm, data, size);

	return err;
}
