#include <stdio.h>
#include <string.h>
#include "arg.h"
#include "log.h"
#include "dep.h"
#include "db.h"
#include "tmpasm_scconfig.h"

static void wtr_set_debug(const char* val)
{
	put("/local/wtruntime/debug", val);
}

static void wtr_set_prefix(const char* val)
{
	put("/local/wtruntime/prefix", val);
}

static int wtr_custom_arg(const char* key, const char* value)
{
	if(0==strcmp(key, "enable-f32"))
	{
		put("/local/wtruntime/enable_f32", value);
		return 1;
	}

	if(0==strcmp(key, "enable-f64"))
	{
		put("/local/wtruntime/enable_f64", value);
		return 1;
	}

	if(0==strcmp(key, "gidtree-nodebits"))
	{
		put("/local/wtruntime/gidtree_nodebits", value);
		return 1;
	}

	return 0;
}

static void wtr_hook_postinit(void)
{
	db_mkdir("/local");
	put("/local/wtruntime/debug", sfalse);
	put("/local/wtruntime/prefix", "/usr");
	put("/local/wtruntime/enable_f32", strue);
	put("/local/wtruntime/enable_f64", strue);
	put("/local/wtruntime/gidtree_nodebits", "0");
}

static int wtr_hook_detect_host(void)
{
	require("fstools/ar", 0, 0);

	return 0;
}

static int wtr_hook_detect_target(void)
{
	char key[64];

	require("cc/ldflags_so", 0, 0);
	require("cc/ldflags_dll", 0, 0);
	require("cc/fpic", 0, 0);
	require("cc/wlrpath", 0, 0);
	require("cc/wloutimplib", 0, 0);
	require("cc/argstd/*", 0, 0);
	require("std/c99/int8_t/presents", 0, 0);
	require("std/c99/int16_t/presents", 0, 0);
	require("std/c99/int32_t/presents", 0, 0);
	require("std/c99/int64_t/presents", 0, 0);
	require("std/c99/uint8_t/presents", 0, 0);
	require("std/c99/uint16_t/presents", 0, 0);
	require("std/c99/uint32_t/presents", 0, 0);
	require("std/c99/uint64_t/presents", 0, 0);
	require("std/c99/uintptr_t/presents", 0, 0);
	require("sys/types/size/*", 0, 0);
	require("sys/ext_dynlib_native", 0, 0);
	require("libs/math/expf/*", 0, 0);

	if(0!=require("sys/types/void_ptr", 0, 0)) { return 1; }
	sprintf(key, "sys/types/size/%s_u_int", get("sys/types/void_ptr"));
	put("/local/wtruntime/uintptr_t", get(key));

	require("cc/cc", 0, 0);
	require("cc/cflags", 0, 0);
	require("cc/argstd/std_c99", 0, 0);

	return 0;
}

static int wtr_hook_generate(const char* root)
{
	int generr = 0;

	fprintf(stderr, "Generating wtruntime/Makefile (%d)\n", generr |= tmpasm(root, "Makefile.in", "Makefile"));
	fprintf(stderr, "Generating wtruntime/config.h (%d)\n", generr |= tmpasm(root, "config.h.in", "config.h"));
	fprintf(stderr, "Generating wtruntime/config_private.h (%d)\n", generr |= tmpasm(root, "config_private.h.in", "config_private.h"));

	return generr;
}
